#include "ymsecure.h"
#include "ymsecure_md5.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <time.h>
#include <openssl/pem.h>
#include <curses.h>

#define LARGE_BUF       4096

/*
	A prime number and a base is hardcoded in the client.
	For now lets use p=7, b=5
	1.	Client initiates a signature message & the first message.
		The first message is displayed as it is and handshake starts.
	2.	If signature is recognized we generate a random number R1 
		and send the result of 
		b^R1 mod p --> Res1
	3.	Other guy generates a random number R2 and sends the result of
		b^R2 mod p --> Res2
	4.	One guy computes, Res2^R1 mod p --> K
	5.	Other computes, Res1^R2 mod p --> K
	6. 	the encryption key is md5(K) --> 16 bytes
	7.	Encryption is repeated XOR with this 16byte key

	once this code is known to work well without issues' i'll 
	start using BIGNUM libs for primes.
*/

struct ymsec_keyinfo gbl_keys[HASHSZ];

/* first check if key is available, if so just encrypt and send message.
   else, start the signature sending sequence, followed by the message.
   if message 2 is not sent as reply then we flag this "to_id" as non-ymshell
   and not encrypt further */

int yahoo_secure_im(char *to, char *msg)
{
	struct ymsec_keyinfo *ymky = ymsecure_findkey(to);

	if(ymky == NULL)
	{
		ymsecure_start_handshake(to);
		return TRY_SECURE;
	}

	if(ymky->flag & YMSECURE_NOYMSHELL)
		return NOSECURE;
	
	ymsecure_encandsend(to, msg, ymky);
	return SECURE;
}

void ymsecure_encandsend(char *to, char *msg, struct ymsec_keyinfo *ymky)
{
	int i, x = 0, len = strlen(msg);
	for(i = 0; i < len; i++)
	{
		msg[i] = msg[i]^ymky->key[x];
		++x;
		if(x == MD5_DIGEST_SZ)	x = 0;
	} 
	len = ymsecure_b64encode((unsigned char *)msg, len);
}	

unsigned char tmpbuf[LARGE_BUF];
int ymsecure_b64encode(unsigned char *msg, int len)
{
	int tlen;
#ifdef YMSDEBUG
	printw("Input string = %s\n", msg);
#endif	
	bzero(tmpbuf, LARGE_BUF);
	tlen = EVP_EncodeBlock(tmpbuf, msg, len);
	if(tlen < 0)
	{
		printw("Error doing base 64 encoding\n");
		exit(1);
	}	
	bcopy(tmpbuf, msg, tlen);
	msg[tlen] = '\0';

#ifdef YMSDEBUG
	printw("Output string = %s\n", msg);
#endif
	return tlen;
}

/* we shall send the sig now */
void ymsecure_start_handshake(char *to)
{
	struct ymsec_keyinfo *ymky = ymsecure_allocinfo(to);
	send_msg_unconditional(YMSHELL_SIGN, to);
	ymky->flag |= YMSECURE_SIGSENT;
	return;
}

struct ymsec_keyinfo *ymsecure_allocinfo(char *id)
{
	int i = ymsecure_hashid(id);
	struct ymsec_keyinfo *newk, *iter;

	newk = (struct ymsec_keyinfo *)malloc(sizeof(struct ymsec_keyinfo));
	bzero(newk, sizeof(struct ymsec_keyinfo));
	strcpy(newk->yid, id);

	iter = &gbl_keys[i];
	while(iter->link != NULL)
		iter = iter->link;
	
	iter->link = newk;
	return newk;
}	

struct ymsec_keyinfo *ymsecure_findkey(char *id)
{
	int i = ymsecure_hashid(id);
	struct ymsec_keyinfo *iter;

	for(iter = gbl_keys[i].link; iter != NULL; iter = iter->link)
	{
		if(strcmp(id, iter->yid) == 0)
			return iter;
	}

	return NULL;
}	

void ymsecure_find_delkey(char *id)
{
	int i = ymsecure_hashid(id);
	struct ymsec_keyinfo *iter, *par;

	par = &gbl_keys[i];
	for(iter = par->link; iter != NULL; iter = iter->link)
	{
		if(strcmp(id, iter->yid) == 0)
		{
			par->link = iter->link;
			free(iter);
#ifdef YMSDEBUG
			printw("Removed key owned by %s\n", id);
#endif		
			return;
		}	
		par = par->link;
	}
	return;
}	

int ymsecure_hashid(char *id)
{
	int i, ret = 0;	
	for(i = 0; i < strlen(id); i++)
		ret = id[i] + (ret << 6) + (ret << 16) - ret;	
	return(ret%HASHSZ);
}	

/* called on receiving a message 
   1. check if its a SIGN, & if it is, alloc new struct & send the special num 
   2. check if its not ymshell, & if its not do nothing 
      
*/

int ymsecure_statemanage(char *umsg, char *from)
{
	struct ymsec_keyinfo *rcvky = ymsecure_findkey(from);
	if(rcvky == NULL)
	{
		if((ymsecure_checksign(umsg, from)) == DONTDISPLAY)
			return DONTDISPLAY;
		return 0;
	}	

	if(rcvky->flag & YMSECURE_SIGSENT)	/* sign sent, check if msg is correct reply */
	{
		ymsecure_checkreplyforsign(rcvky, umsg);
		if(rcvky->flag & YMSECURE_NOYMSHELL)
			return 0;
		return DONTDISPLAY;
	}

	if(rcvky->flag & YMSECURE_SECRETSENT) 
	{
		ymsecure_checkreplyforsecret(rcvky, umsg);
		if(rcvky->flag & YMSECURE_NOYMSHELL)
			return 0;
		return DONTDISPLAY;
	}	

	if(rcvky->flag & YMSECURE_KEYEXISTS)
	{
		ymsecure_decrypt(rcvky, umsg);
		return 0;
	}	

	if(rcvky->flag & YMSECURE_NOYMSHELL)
		return 0;

	return 0;
}	

extern unsigned char tmpbuf[LARGE_BUF];
void ymsecure_decrypt(struct ymsec_keyinfo *ymky, char *msg)
{
	int i, x = 0, len;

	bzero(tmpbuf, LARGE_BUF);
	len = EVP_DecodeBlock(tmpbuf, (unsigned char *)msg, strlen(msg));
	if(len < 0)
	{
		printf("Error\n");
		exit(1);
	}
	bcopy(tmpbuf, msg, len);
	msg[len] = '\0';

	for(i = 0; i < strlen(msg); i++)
	{
		msg[i] = msg[i]^ymky->key[x];
		++x;
		if(x == MD5_DIGEST_SZ)	x = 0;
	}
}

void ymsecure_checkreplyforsecret(struct ymsec_keyinfo *ky, char *umsg)
{
	if(strncmp(umsg, HNDSHK_REPLY_2, strlen(HNDSHK_REPLY_2)) == 0)
	{
		ky->flag &= ~YMSECURE_SECRETSENT;
		ky->recvnum = atoi(&umsg[strlen(HNDSHK_REPLY_2)]);
		ymsecure_calckey(ky);
	}	
	else
	{
		ky->flag |= YMSECURE_NOYMSHELL;
	}	
	return;
}	

void ymsecure_checkreplyforsign(struct ymsec_keyinfo *ky, char *umsg)
{
	if(strncmp(umsg, HNDSHK_REPLY_1, strlen(HNDSHK_REPLY_1)) == 0)
	{
		ky->flag &= ~YMSECURE_SIGSENT;
		ky->recvnum = atoi(&umsg[strlen(HNDSHK_REPLY_1)]);
		ymsecure_sendfinalreply(ky, HNDSHK_REPLY_2);
		ymsecure_calckey(ky);
	}	
	else
	{
		ky->flag |= YMSECURE_NOYMSHELL;
	}
	return;
}	

int ymsecure_checksign(char *umsg, char *from)
{
	struct ymsec_keyinfo *newk;

	if(strncmp(umsg,  YMSHELL_SIGN, strlen(YMSHELL_SIGN)) == 0)
	{	
		newk = ymsecure_allocinfo(from);
		ymsecure_sendsignreply(newk, HNDSHK_REPLY_1);
		return DONTDISPLAY;
	}
	else
	{
#ifdef YMSDEBUG
		printw("Not YMSHELL or secure not enabled\n");
#endif		
	}
	return 0;	
}	

void ymsecure_sendsignreply(struct ymsec_keyinfo *ky, char *reply)
{
	double res;
	char tbuf[MAX_BUF];

	srand((int )time(NULL));
	ky->rnd = rand()%RAND_MX;
	res = pow((double) PRIMV, (double) ky->rnd);
	res = (long long int)res%PRIME;
	sprintf(tbuf, "%s%d\n", reply, (int )res);
	send_msg_unconditional(tbuf, ky->yid);
	ky->flag |= YMSECURE_SECRETSENT;
}

void ymsecure_sendfinalreply(struct ymsec_keyinfo *ky, char *reply)
{
	double res;
	char tbuf[MAX_BUF];
	
	srand((int )time(NULL));
	ky->rnd = rand()%RAND_MX;
	res = pow((double) PRIMV, (double )ky->rnd);
	res = (long long int)res%PRIME;
	sprintf(tbuf, "%s%d\n", reply, (int ) res);
	send_msg_unconditional(tbuf, ky->yid);
}	

void ymsecure_calckey(struct ymsec_keyinfo *ky)
{
	double k;
	k = pow((double)ky->recvnum, (double)ky->rnd);
	k = (long long int)k%PRIME;

	ky->sesskey = k;
	ymsecure_mdkeymake(ky);
	ky->flag |= YMSECURE_KEYEXISTS;
	ky->flag &= ~YMSECURE_NOYMSHELL;
#ifdef YMSDEBUG
	printw("--> Session key = %d\n", ky->sesskey);
	fflush(stdout);
#endif	
}	
	
void ymsecure_mdkeymake(struct ymsec_keyinfo *ky)
{
	char tmp[MAX_BUF];
	MD5_CTX ctx;
	int n;

	bzero(&ctx, sizeof(MD5_CTX));
	n = sprintf((char *) tmp, "%d", ky->sesskey);
	ky->key = (unsigned char *) malloc (MD5_DIGEST_SZ);
	MD5Init(&ctx);
	MD5Update(&ctx, tmp, strlen(tmp));
	MD5Final(&ctx);
	bcopy(ctx.digest, ky->key, MD5_DIGEST_SZ);
#ifdef YMSDEBUG
	printw("--> Enc key");
	for(n = 0; n < 16; n++)
		printw("%d:", ky->key[n]);
	printw("\n");	
	fflush(stdout);
#endif	
}

void ymsecure_update_state(char *who)
{
	ymsecure_find_delkey(who);
	return;
}	
