/*
 * sample yahoo client based on libyahoo2
 *
 * libyahoo2 is available at http://libyahoo2.sourceforge.net/
 *
 * $Revision: 1.72 $
 * $Date: 2004/01/16 05:38:46 $
 * 
 * Copyright (C) 2002-2004, Philip S Tellis <philip.tellis AT gmx.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* ymshell.c, edited from the original sample_client.c of libyahoo2 */
/* Author: Tejus A.G , Ranveer */

#include "ymshell.h"
#include "ymgroup.h"
#include "ymcallback.h"
#include "ymcontext.h"
#include "ymhistory.h"
#ifdef SECURE_CHAT
#include "ymsecure.h"
#endif

char mbuf    [LARGE_BUF];	/* messaging buffer 		*/
char st	     [MAX_BUF];		/* status message buffer 	*/
char armsg   [MAX_BUF];		/* auto reply message buffer 	*/
char to_id   [MAX_BUF];		/* to id 			*/
char from_id [MAX_BUF];		/* from id			*/
int ix;				/* running index into mbuf	*/

ymshell_vals 	  	cfg; 		/* config/running parameters 	*/
ymshell_msglist		hdr;		/* saved history		*/
struct YGroupList 	*buddygroups;	/* buddy group list 		*/
FILE              	*logfd;		/* log file desc		*/
time_t            	last;		/* time tracker 		*/
yahoo_local_account 	*ylad;		/* my account info		*/
YList 			*buddies;	/* buddy list 			*/
YList			*tmpbuddies;	/* buddy list			*/
struct shortcuts 	sc;		/* shortcuts configured		*/
extern char **environ;
extern void run_config_wizard(char *);

char *help_string="\n**************************************************************************"
		  "\n                               YMSHELL 1.1 			       "
		  "\n		           Authors: Tejus A.G, Ranveer                         "
		  "\n**************************************************************************"
		  "\nCommands\n"
		  "--------\n"
		  "ls           : List Buddies          Displays indexed list of online buddies\n"
		  "ls -l        : Get Buddies           Lists all buddies (online & offline )\n"
		  "talk         : Message Buddy         Initiates a message to buddy \n"
		  "talk -i      : Message Index         Initiates a message index (index in ls list)\n"
		  "exit         : Logout                Logs off yahoo\n"
		  "bell         : Notification          Enable/Disable message arrival notification\n"
		  "avail        : Set avail             Sets status to AVAILABLE\n"
		  "invis        : Set invis             Sets status to INVISIBLE\n"
		  "status       : Status Message        Sets a status message\n"
		  "state        : Show State            Shows current yahoo status\n"
		  "bash,sh      : Run shell             Switchs to shell, typing exit brings back to ymshell\n"
		  "ps           : Show contexts         Shows list of chat contexts (a.k.a tabs\n"
		  "cls          : Clears Screen         Clear screen\n"
		  "touch        : Add a buddy           Adds a buddy to group ymshell_grp\n"
		  "rm           : Remove buddy          Removes named buddy from friend list\n"
		  "auto         : Auto reply            Set a auto-reply message\n"
		  "info         : Help                  Displays this help\n"
		  "secure       : Secure                Enable secure chat (experimental)\n"
		  "reap <id>    : Reap context          Removes a context identified by id\n"
		  "**************************************************************************\n";

yahoo_idlabel yahoo_status_codes[] = 
{
	{YAHOO_STATUS_AVAILABLE, 	"AVAILABLE"	},
  	{YAHOO_STATUS_BRB, 		"BRB"		},
 	{YAHOO_STATUS_BUSY, 		"BUSY"		},
  	{YAHOO_STATUS_NOTATHOME, 	"NOT HOME"	},
  	{YAHOO_STATUS_NOTATDESK, 	"NOT AT DESK"	},
  	{YAHOO_STATUS_NOTINOFFICE, 	"NOT IN OFFICE" },
  	{YAHOO_STATUS_ONPHONE,		"ON PHONE"	},
  	{YAHOO_STATUS_ONVACATION, 	"ON VACATION"	},
  	{YAHOO_STATUS_OUTTOLUNCH, 	"OUT TO LUNCH"  },
  	{YAHOO_STATUS_STEPPEDOUT, 	"STEPPED OUT"	},
  	{YAHOO_STATUS_INVISIBLE, 	"INVISIBLE"	},
  	{YAHOO_STATUS_IDLE, 		"IDLE"		},
  	{YAHOO_STATUS_OFFLINE, 		"OFFLINE"	},
  	{YAHOO_STATUS_CUSTOM, 		"[CUSTOM]"	},
  	{YAHOO_STATUS_NOTIFY, 		"NOTIFY"	},
  	{0, 				NULL		}
};

/******************* UTILITY FUNCTIONS ********************/

#define COMMAND_BUF_EMPTY	(ix == 0)

#define remove_html_formatting(str) do {\
	int _l = strlen(str), i=0, j=0;\
	while(i <= _l)\
	{ \
	retest:\
		if(str[i] == '\0')	break; \
		if(str[i] == '<')\
		{\
			while(str[i] != '>')\
			{\
				i=i+1; /* skip */\
			}\
			++i;\
			goto retest;\
		}\
		str[j]=str[i];\
		j=j+1;\
		i=i+1;\
	}\
	str[j]='\0';\
}while(0)\

void ctrl_z_handler()
{
	def_prog_mode();
	endwin();
	system("echo \"Type exit to return to ymShell :) \"");
	system("bash");
	reset_prog_mode();
	touchwin(stdscr);
	wrefresh(stdscr);
}	

void drop_prompt()
{
	if(!cfg.message_mode)
		printw("ymShell $");
	else
	{
		wattron(stdscr, COLOR_PAIR(2));
		printw("%s << ", to_id);
		wattroff(stdscr, COLOR_PAIR(2));
	}	
}		
		
void	read_yahooconf()
{
	char buf[MAX_BUF], *tok;
	FILE *fp=fopen(mbuf, "r"); /* mbuf was set to conf file in is_first_run */
	cfg.init_state = YAHOO_STATUS_AVAILABLE;
	cfg.max_hist = 20;
	cfg.secure = 0;	
	sc.send_cycle = 's';
	sc.recv_cycle = 'r';

	if(fp == NULL)	return;
	while((fscanf(fp, "%s", buf)) != EOF)
	{	
		tok = strtok(buf,"=");
		if(tok[0] == 'Y' && tok[1] == 'A')
		{
			tok=strtok(NULL,"=");
			strcpy(ylad->yahoo_id, tok);
		}
		if(tok[0] == 'P' && tok[1] == 'W')
		{
			tok=strtok(NULL,"=");
			strcpy(ylad->password, tok);
		}
		if(tok[0] == 'S' && tok[1] == 'T')
		{
			tok=strtok(NULL,"=");
			if(strcasecmp(tok,"invisible") == 0)
				cfg.init_state = YAHOO_STATUS_INVISIBLE;
			if(strcasecmp(tok,"available") == 0)
				cfg.init_state = YAHOO_STATUS_AVAILABLE;
		}
		if(strncmp(tok, "PORT", strlen("PORT")) == 0)
		{
			tok=strtok(NULL,"=");
			cfg.port_to_use=atoi(tok);
		}
		if(strncmp(tok, "HIST", strlen("HIST")) == 0)
		{
			tok=strtok(NULL,"=");
			cfg.max_hist=atoi(tok);
		}	
		if(strncmp(tok, "RECVSC", strlen("RECVSC")) == 0)
		{
			tok=strtok(NULL,"=");
			sc.recv_cycle = tok[0];
		}	
		if(strncmp(tok, "SENDSC", strlen("SENDSC")) == 0)
		{
			tok=strtok(NULL,"=");
			sc.send_cycle = tok[0];
		}	
	}	
}
	
char *get_buddy_name (char *yid)
{
	YList *b;
	for (b = buddies; b; b = b->next)
	{
		yahoo_account *ya = b->data;
		if (!strcmp (yid, ya->yahoo_id))
			return ya->name && *ya->name ? ya->name : ya->yahoo_id;
    	}
	return yid;
}

char *yahoo_status_code (enum yahoo_status s)
{
	int i;
	for (i = 0; yahoo_status_codes[i].label; i++)
		if (yahoo_status_codes[i].id == s)
			return yahoo_status_codes[i].label;
	return "Unknown";
}

void yahoo_ping_timeout_callback (int sig)
{
	fprintf (logfd, "PNG: sending a keep alive msg\n");
	yahoo_keepalive (ylad->id);
}

/*************** YMSHELL specific functions ****************/

char *find_ingroup(char *bud)
{
	YGroupList *itr=buddygroups;
	if(!buddygroups) return NULL;
	yahoo_account_list *ya_list;
	yahoo_account *ya;

	do {
		ya_list = itr->ya_list;
		do {
			ya = ya_list->ya;
			if(strcmp(bud, ya->yahoo_id) == 0)
				return itr->name;
			ya_list=ya_list->prev;
		}while(ya_list!=itr->ya_list);
		itr=itr->prev;
	}while(itr!=buddygroups);

	return NULL;
}	

void ymshell_add_buddy(char *bud)
{
	yahoo_add_buddy(ylad->id, bud, "ymshell_grp", "Hi . . .!");
	printw("Adding buddy %s to \"ymshell_grp\"...\n", bud);
	fflush(stdout);
	yahoo_get_list(ylad->id);
	yahoo_refresh(ylad->id);
}

void ymshell_remove_buddy(char *bud)
{
	char *grp = find_ingroup(bud);
	if(grp == NULL)
	{
		printw("No such buddy\n");
		goto out;
	}	

	yahoo_remove_buddy(ylad->id, bud, grp);
	printw("Removing buddy %s from %s...\n", bud, grp);
out:	
	fflush(stdout);
	yahoo_get_list(ylad->id);
	yahoo_refresh(ylad->id);
}	

/* Send a kde notification on some event using kdialog.
   When no kdialog is present the function does nothing. */

void ymshell_notification(char *str, int flag)
{
	time_t new;
	char nbuf[MAX_BUF];
	
	if(cfg.kde_pres == 0)	return;

	if(cfg.notice)
	{
		if(flag)	goto doit;
		new = time(NULL);
		if(new - last > NOTIFY_GAP)
		{
			last = new;
		doit:
			strcpy(nbuf, "kdialog --passivepopup ");
			strcat(nbuf, str);
			strcat(nbuf, " 1 ");
			strcat(nbuf, " 2> /dev/null > /dev/null &");
			system(nbuf);
		}
	}	

}
void ymshell_show_state()
{
	printw("You are currently %s\n", st);
	if(ylad->msg && ylad->status == YAHOO_STATUS_CUSTOM)
		printw("Your status message = %s\n", ylad->msg);
	fflush(stdout);
}

void ymshell_show_context() {
	int i;
	for(i=0; i<num_context; ++i)
		printw("context:%i %s\n", i+1, context[i]);
	printw("\n");

}

void ymshell_set_statemsg(char *msg)
{
	if(msg == NULL)	return;
	if(ylad->msg)	free(ylad->msg);
	ylad->msg = (char *)malloc(strlen(msg) + 1);
	strcpy(ylad->msg, msg);
	ylad->status = YAHOO_STATUS_CUSTOM;
	yahoo_set_away (ylad->id, YAHOO_STATUS_CUSTOM, ylad->msg, 0);
	strcpy(st, "AVAILABLE");
}

/* Cmd $ar off, switches off auto replies */

void ymshell_set_autoreply(char *msg)
{
	if(msg == NULL)	return;
	if(!strcasecmp(msg, "off"))
	{
		cfg.ar = 0;
	}	
	else
	{
		cfg.ar = 1;
		strcpy(armsg, "AUTO-GENERATED-MSG: ");
		strcat(armsg, msg);	
	}	
}	

void ymshell_notify_mailreceipt(char *from, char *subj)
{
}	

void ymshell_message_id(char *who)
{
	if(!who)	return;
	cfg.message_mode = 1;
	strcpy(to_id, who); 	
}

void ymshell_message_index(char *to_who)
{
	int i, j;
 	YGroupList *b = buddygroups;
	yahoo_account_list *ya_list;
	if(!to_who) 	return;
	sscanf(to_who, "%u.%u", &i, &j);
	do {
		if(b->id == i)
			break;
		b=b->prev;
	} while(b!=buddygroups);
	ya_list = b->ya_list;
	do {
		if(ya_list->id == j)
			break;
		ya_list = ya_list->next;
	} while(ya_list!=b->ya_list);
	yahoo_account *ya = ya_list->ya;

	strcpy(to_id, ya->yahoo_id);
	cfg.message_mode = 1;
	return;
}

void ymshell_message_group(int group_index, char *msg, char spam) {
	int i = group_index;
	char *utf;
 	YGroupList *b = buddygroups;
	yahoo_account_list *ya_list;
	yahoo_account *ya;
	if(!group_index) 	return;
	do {
		if(b->id == i)
			break;
		b=b->prev;
	} while(b!=buddygroups);
	wattron(stdscr, COLOR_PAIR(2));
	if(spam)
		printw("%s << %s\n", b->name, msg);
	else
		printw("[%s] << %s\n", b->name, msg);
	wattroff(stdscr, COLOR_PAIR(2));
	utf = y_str_to_utf8(msg);
	ya_list = b->ya_list;
	do {
		ya = ya_list->ya;
		if (spam || ya->status != YAHOO_STATUS_OFFLINE) {
			yahoo_send_im(ylad->id, NULL, ya->yahoo_id, utf, 1, 0);
		}
		ya_list = ya_list->next;
	} while(ya_list!=b->ya_list);
	FREE(utf);

	fflush(stdout);
	return;
}

void ymshell_message_all(char *msg, char spam) {
}

void ymshell_display_help()
{
	if(cfg.connected)
	{
		erase();
		wattron(stdscr, COLOR_PAIR(2) | A_BOLD);
		printw("%s", help_string);
		wattroff(stdscr, COLOR_PAIR(2) | A_BOLD);
	}
	else
		printf("%s", help_string);
	fflush(stdout);
		
}

void ymshell_change_state(int seen)
{
	if(seen == 1)
	{
		printw("Setting to available...\n");
		ylad->status = YAHOO_STATUS_AVAILABLE;
		yahoo_set_away(ylad->id, YAHOO_STATUS_AVAILABLE, NULL,0);
		strcpy(st, "AVAILABLE");
	}
	if(seen == 0)
	{
		printw("Setting to invisible...\n");
		ylad->status = YAHOO_STATUS_INVISIBLE;
		yahoo_set_away(ylad->id, YAHOO_STATUS_INVISIBLE, NULL,0);
		strcpy(st, "INVISIBLE");
	}
	fflush(stdout);
}


void ymshell_list_groups() {
	if(!buddygroups)return;
	YGroupList *itr=buddygroups;
	erase();
	printw("Groups\n---------------\n");
	do {
		wattron(stdscr, COLOR_PAIR(3) | A_BOLD);
		printw("| %u | %s\n", itr->id, itr->name);
		wattroff(stdscr, COLOR_PAIR(3) | A_BOLD);
		itr=itr->prev;
	}while(itr!=buddygroups);
	printw("\n");
	fflush(stdout);
	return;
}

/* TODO: listing names using realname or nick name */

void ymshell_listonline_buddies(char *tok)
{
	int ctr, notmt = 0;
	char buf[MAX_BUF];

	if(!buddygroups)return;
	YGroupList *itr=buddygroups;
	yahoo_account_list *ya_list, *ya_list_tmp;
	yahoo_account *ya;

	erase();
	printw("Online Buddies\n---------------\n");
	do {
		if(tok && (atoi(tok) != itr->id)) {
			itr = itr->prev;
			continue;
		}
		ya_list = itr->ya_list;
		ya_list_tmp = ya_list;
		
		do {
			ya = ya_list->ya;
			if(ya->status != YAHOO_STATUS_OFFLINE)
			{	
				notmt = 1;
				break;
			}
			ya_list = ya_list->prev;
		}while(ya_list!=itr->ya_list);
		ya_list = ya_list_tmp;
		if(notmt)
		{
			wattron(stdscr, COLOR_PAIR(3) | A_BOLD);
			printw("| %u | %s\n", itr->id, itr->name);
			wattroff(stdscr, COLOR_PAIR(3) | A_BOLD);
			do {
				wattron(stdscr, COLOR_PAIR(2));
				ya = ya_list->ya;
				if (ya->status != YAHOO_STATUS_OFFLINE)
					{
						sprintf(&buf[0], "| %d.%-2d | %s", itr->id, ya_list->id, ya->yahoo_id);
						printw("%s", buf);
						ctr=strlen(buf);
						while(ctr != 40)	++ctr, printw(" ");
						printw("%s", yahoo_status_code (ya->status));
						if(ya->status == YAHOO_STATUS_CUSTOM && ya->msg)
							printw(" - %s\n", ya->msg);
						else
							printw("\n");
					}
				ya_list=ya_list->prev;
				wattroff(stdscr, COLOR_PAIR(2));
			}while(ya_list!=itr->ya_list);
		}	
		if(notmt)
			printw("\n");
		itr=itr->prev;
		notmt = 0;
	}while(itr!=buddygroups);
	printw("\n");
	fflush(stdout);
	return;
}

void ymshell_listall_buddies(char *tok)
{
	int ctr, yy, xx;
	int cury = 1;
	char buf[128], ch;
	YGroupList *itr=buddygroups;
	if(!buddygroups)return;
	yahoo_account_list *ya_list;
	yahoo_account *ya;

	erase();
	getmaxyx(stdscr, yy, xx);
	printw("Buddy List\n----------\n");
	cury+=2;
	do {
		if(tok && (atoi(tok) != itr->id)) {
			itr=itr->prev;
			continue;
		}
		wattron(stdscr, COLOR_PAIR(3) | A_BOLD);
		printw("| %u | %s\n", itr->id, itr->name);
		cury+=1;
		wattroff(stdscr, COLOR_PAIR(3) | A_BOLD);
		ya_list = itr->ya_list;
		do {
			if(cury >= yy)
				scanf("%c",&ch);
			wattron(stdscr, COLOR_PAIR(2));
			ya = ya_list->ya;
			sprintf(&buf[0], "| %d.%-2d |   %s", itr->id, ya_list->id, ya->yahoo_id);
			printw("%s", buf);
			ctr=strlen(buf);
			while(ctr != 40)	++ctr, printw(" ");
			printw("%s", yahoo_status_code (ya->status));
			if(ya->status == YAHOO_STATUS_CUSTOM && ya->msg)
				printw(" - %s\n", ya->msg);
			else
				printw("\n"); 
			ya_list=ya_list->prev;
			wattroff(stdscr, COLOR_PAIR(2));
			cury = cury + 1;
			fflush(stdout);
			touchwin(stdscr);
			wrefresh(stdscr);
		}while(ya_list!=itr->ya_list);
		printw("\n");
		cury+=1;
		itr=itr->prev;
	}while(itr!=buddygroups);
	fflush(stdout);
	return;
}

void ymshell_clear()
{
	erase();
	touchwin(stdscr);
	wrefresh(stdscr);
}

void ymshell_remove_context(char *to)
{
	/* Ranveer: implement code to remove 'to' from context list */  
	/* Tejus:   implement freeing of keyinfo for secure
		    and free saved message history  		    */
}  

void ymshell_restore_history(char *to)
{
	ymshell_msglist *cr = hdr.nxt;
	ymshell_msgs *iter;
	int attr;

	erase();
	while(cr != NULL)
	{
		if((strcmp(cr->id, to)) == 0)
			break;
		cr = cr->nxt;	
	}
	if(cr == NULL)	return;
	iter = cr->msgs;
	while(iter != NULL)
	{
		if(iter->type == SENT)	
		{
			attr = COLOR_PAIR(2);
			wattron(stdscr, attr);
			printw("%s << ", to_id);
		}	
		if(iter->type == RECD) 
		{
			attr = COLOR_PAIR(3);
			wattron(stdscr, attr);
			printw("%s >> ", to_id);
		}	
		wattroff(stdscr, attr);
		printw("%s\n", iter->buf);
		iter = iter->nxt;
	}
	touchwin(stdscr);
}	

#define MAKE_BUF(_x, _st, _t)				do {    \
	_x = (ymshell_msgs *) malloc(sizeof(ymshell_msgs)); 	\
	_x->buf = (char *)malloc(strlen(_st)+1);		\
	strcpy(_x->buf, _st);					\
	_x->type = _t;						\
	_x->nxt = NULL;						\
} while	(0)

void ymshell_buffer_msg(ymshell_msglist *itr, char *msg, int type)
{
	ymshell_msgs *cr;
	ymshell_msgs *bf;

	if(itr->num == 0)
	{
		MAKE_BUF(itr->msgs, msg, type);
		itr->num +=1;
		return;
	}

	cr = itr->msgs;
	while(cr->nxt != NULL)	 cr = cr->nxt;
	MAKE_BUF(cr->nxt, msg, type);
	itr->num +=1;

	if(itr->num == cfg.max_hist)
	{
		bf = itr->msgs;
		itr->msgs = itr->msgs->nxt;
		free(bf->buf);
		free(bf);
		itr->num -=1;
	}	
	return;
}

void ymshell_store_message(char *msg, char *to, int type)
{
	/* remember upto 10 messages for the users */
	/* this buffers will be free'd if 'reap' command is used on that user */
	/* the history gets displayed whenever context switch is performed */

	ymshell_msglist *itr = &hdr;

	itr = itr->nxt;
	while(itr != NULL)
	{
		if(strcmp(itr->id, to) == 0)
			goto addbuf;
		itr = itr->nxt;
	}
	itr = (ymshell_msglist *)malloc(sizeof(ymshell_msglist));
	itr->num = 0;
	itr->msgs = NULL;
	strcpy(itr->id, to);
	itr->nxt = hdr.nxt;
	hdr.nxt = itr;
addbuf:
	ymshell_buffer_msg(itr, msg, type);
}

void ymshell_nosuch_command(char *cmd)
{
	wattron(stdscr, COLOR_PAIR(2) | A_BOLD);
	printw("\nymShell: %s: command not found. type info for help\n\n", cmd);
	wattroff(stdscr, COLOR_PAIR(2) | A_BOLD);
}	

void send_msg_unconditional(char *msg, char *to)
{
	char *utf;
	utf = y_str_to_utf8(msg);
	yahoo_send_im(ylad->id, NULL, to, utf, 1, 0);
	free(utf);
}	

void send_msg(char *msg)
{
	char *utf;

#ifdef SECURE_CHAT	
	if(cfg.secure)
	{
		int i;
		i = yahoo_secure_im(to_id, msg);
		if(i == NOSECURE || i == TRY_SECURE)	goto sendnow;
		ymshell_store_message(msg, to_id, SENT);
		yahoo_send_im(ylad->id, NULL, to_id, msg, 0, 0);
	}
#else
	if(0)	{ goto sendnow; }
#endif	
	else
	{
sendnow:	
		utf = y_str_to_utf8(msg);
		ymshell_store_message(msg, to_id, SENT);
		yahoo_send_im(ylad->id, NULL, to_id, utf, 1, 0);
		FREE(utf);
	}
	ymupdate_context();
	ymupdate_cl(send_cl, to_id);
	cfg.reset_flag=1;
}
void ymshell_send_autoreply(char *who)
{
	send_msg_unconditional(armsg, who);
}	

void ymshell_disp_got_message(char *id, char *buf, int msgtype)
{
	int x, y, i;
	int attrs;

	switch(msgtype)
	{
		case INCTX:
			attrs = COLOR_PAIR(3);
			break;
		case OFFLINE:
			attrs = COLOR_PAIR(1) | A_BOLD;
			break;
		case INCTX_DIFFUSER:
			attrs = COLOR_PAIR(2) | A_BOLD;
			break;
		case NOCTX:
			attrs = COLOR_PAIR(2) | A_BOLD;
			break;
	}		
	
	if(COMMAND_BUF_EMPTY)	// buffer empty 
	{
		wattron(stdscr, attrs);
 		printw("\r%s", id);
		wattroff(stdscr, attrs);
 		printw("%s", buf);
		drop_prompt();
	}	
	else
	{
		getyx(stdscr, y, x);	
		move(y, 0);
		clrtoeol();
		wattron(stdscr, attrs);
 		printw("\r%s", id);
		wattroff(stdscr, attrs);
 		printw("%s", buf);
		drop_prompt();
		for(i = 0; i < ix; i++)
			printw("%c", mbuf[i]);
	}		
	touchwin(stdscr);
	wrefresh(stdscr);
}	

/****************************************************************/
/********************* IMP ext_yahoo funcs **********************/

/* this is called when a message is received 	*/
/* id is conn ID				*/
/* who is sender				*/
/* msg is the message data			*/
/* tm is timestamp(for offline messages)	*/
/* stat is no idea what				*/
/* utf8 is flag for UTF encoded messages	*/

void ext_yahoo_got_im (int id, const char *me, const char *who, const char *msg, long tm, int stat, int utf8)
{
	char *umsg = msg, idc[MAX_BUF], buf[LARGE_BUF], tb[MAX_BUF];
	int type;

	if (stat == 2)	return;
	if (!msg)	return;
 	if (utf8)       umsg = y_utf8_to_str (msg);
#ifdef SECURE_CHAT
	if (cfg.secure)
	{
		/* dont display the behind the scenes messages */
		if((ymsecure_statemanage(umsg, who)) == DONTDISPLAY)
			return;
	}	
#endif	
  	who = get_buddy_name (who);
	strcpy(from_id, who);
	ymupdate_cl(recv_cl, from_id);

	remove_html_formatting(umsg);			/* some times yahoo sends html data */
	if(tm)						/* in case of web messenger users */
	{
		sprintf(idc, "\n[ Offline message from %s >> ] ", who);
		sprintf(buf, "%s", umsg);
		type = OFFLINE;
	}	
  	else
    	{
		ymshell_store_message(umsg, who, RECD);	/* store upto 5 recent messages */
		if(cfg.message_mode)
		{
			if((strcmp(who, to_id)) == 0)
			{
				sprintf(idc, "%s >> ", who);
				sprintf(buf,"%s\n", umsg);
				type = INCTX;
			}	
			else
			{
				sprintf(idc, "[%s] >> ", who);
				sprintf(buf, "%s\n", umsg);
				type = INCTX_DIFFUSER;
			}	
		}
		else
		{
			sprintf(idc, "[%s] >> ", who);
			sprintf(buf, "%s\n", umsg);
			type = NOCTX;
		}	
	}
	ymshell_disp_got_message(idc, buf, type);
	if(cfg.ar)	ymshell_send_autoreply(who);
	sprintf(tb, "\"Message from %s\"", who);
	cfg.reset_flag=1;
	ymshell_notification(tb, 0);
  	if (utf8)	FREE (umsg);
	fflush(stdout);
}

void ext_yahoo_chat_message (int id, const char *me, const char *who, const char *room, const char *msg, int msgtype, int utf8)
{
}

/* Received whenever status of a buddy changes.
 * First time this is called for all non-offline buddies.
 * and after that whenever the status changes.
 */

void ext_yahoo_status_changed (int id, const char *who, int stat, const char *msg, int away, int idle, int mobile)
{
	yahoo_account *ya = NULL;
	YList *b;
	char tb[MAX_BUF];
	
	for (b = buddies; b; b = b->next)
	{
		if (!strcmp (((yahoo_account *) b->data)->yahoo_id, who))
		{
			ya = b->data;
			break;
		}
	}
  	if (ya)
	{
		if(ya->status == YAHOO_STATUS_OFFLINE && stat == YAHOO_STATUS_AVAILABLE)
		{
			sprintf(tb, "\"%s is now ONLINE\"", who);
			ymshell_notification(tb, 1);
		}
		if(ya->status == YAHOO_STATUS_AVAILABLE && stat == YAHOO_STATUS_OFFLINE)	
		{
			sprintf(tb, "\"%s is now OFFLINE\"", who);
			ymshell_notification(tb, 1);
			/* remove key info of this person */
		#ifdef SECURE_CHAT	
			if(cfg.secure)
				ymsecure_update_state(who);
		#endif		
		}	
		ya->status = stat;
		ya->away = away;
		if (msg)
		{	
	  		FREE (ya->msg);
	  		ya->msg = strdup (msg);
		}
    	}
}

/* After loggin in we receive the entire list of our
 * buddies. At this time all are considered to be
 * offline. a status changed is received for all
 * online buddies */

void ext_yahoo_got_buddies (int id, YList * buds)
{
  	for (; buds; buds = buds->next)
    	{
      		yahoo_account *ya, *sya;
      		struct yahoo_buddy *bud = buds->data;
			ya = y_new0 (yahoo_account, 1);
      		strncpy (ya->yahoo_id, bud->id, 255);
			ya->name[0] = '\0';
      		strncpy (ya->group, bud->group, 255);
			if(buddygroups) {
				sya = search_glist(bud->id);
				if(sya)
					ya->status = sya->status;
				else
					ya->status = YAHOO_STATUS_OFFLINE;
			} else
				ya->status = YAHOO_STATUS_OFFLINE;
      		tmpbuddies = y_list_append (tmpbuddies, ya);
    	}
	while (buddies)
    	{
			YList *tmp;
      		FREE (buddies->data);
			tmp = buddies;
      		buddies = buddies->next;
      		FREE (tmp);
    	}
	buddies = tmpbuddies;
	tmpbuddies = NULL;
	if(buddygroups)
		free_glist();
	mk_glist();
}

void yahoo_logout ()
{
	if (ylad->id <= 0)  	return;
  	yahoo_logoff (ylad->id);
	yahoo_close (ylad->id);
	ylad->status = YAHOO_STATUS_OFFLINE;
  	ylad->id = 0;
  	cfg.poll_loop = 0;
	endwin();
  	printf("\nLogged out.\n");
}

void ext_yahoo_login (yahoo_local_account * ylad)
{

	ylad->id = yahoo_init_with_attributes(ylad->yahoo_id, ylad->password, "pager_port", cfg.port_to_use, NULL);
	ylad->status = YAHOO_STATUS_OFFLINE;
	printf("Setting %s mode\n", yahoo_status_code(cfg.init_state));
	printf("Enabling kdialog notification if available\n");
	strcpy(st, yahoo_status_code(cfg.init_state));
	ylad->status = cfg.init_state;
	yahoo_login (ylad->id, cfg.init_state);
}

void ext_yahoo_login_response (int id, int succ, const char *url)
{
	switch(succ)
	{
	case YAHOO_LOGIN_OK:
  				ylad->status = yahoo_current_status (id);
				cfg.connected = 1;
				init_rest();
				ymshell_clear();
				printw("\nLogged in ...\n");
				drop_prompt();
				fflush(stdout);
			        return;
	case YAHOO_LOGIN_UNAME:
				printf("Could not login: UserName not recognized\n");
				fflush(stdout);
				break;
  	case YAHOO_LOGIN_PASSWD:
				printf("Could not login: Password incorrect\n");
				fflush(stdout);
				break;
  	case YAHOO_LOGIN_LOCK:
				printf("Could not login: Account has been locked\n");
				fflush(stdout);
				break;
	case YAHOO_LOGIN_DUPL:
      				printf("You have been logged due to a duplicate login.\n");
				fflush(stdout);
				break;
	case YAHOO_LOGIN_SOCK:
				printf("Login failed: The server closed the socket.\n");
				fflush(stdout);
				break;
	default:
				return;
	}
  	ylad->status = YAHOO_STATUS_OFFLINE;
  	yahoo_logout ();
	printw("AT LOGOUT\n\n\n");
	sleep(100);
	cfg.poll_loop = 0;
}

void ext_yahoo_error (int id, const char *err, int fatal, int num)
{
	fprintf (stdout, "Yahoo Error: %s\n", err);
	if (fatal)    yahoo_logout ();
}

int ext_yahoo_connect (const char *host, int port)
{
	struct sockaddr_in serv_addr;
	struct hostent *server;
	char last_host[256];
	int servfd;
	char **p;

	if (last_host[0] || strcasecmp (last_host, host) != 0)
	{
      		if (!(server = gethostbyname (host)))
			return -1;
		strncpy (last_host, host, 255);
    	}

  	if ((servfd = socket (AF_INET, SOCK_STREAM, 0)) < 0)
      		return -1;

  	for (p = server->h_addr_list; *p; p++)
    	{
      		memset (&serv_addr, 0, sizeof (serv_addr));
      		serv_addr.sin_family = AF_INET;
      		memcpy (&serv_addr.sin_addr.s_addr, *p, server->h_length);
      		serv_addr.sin_port = htons (port);

      		if (connect (servfd, (struct sockaddr *) &serv_addr, sizeof (serv_addr)) == -1)
		{
	  		if (errno != ECONNREFUSED && errno != ETIMEDOUT && errno != ENETUNREACH)
	    		{
	      			break;
	    		}
		}
      		else
		{
	  		return servfd;
		}
    	}

 	close (servfd);
  	return -1;
}

/*************************************
 * Callback handling code starts here
 */
YList *connections = NULL;
struct _conn
{
	int tag;
  	int fd;
  	int id;
  	yahoo_input_condition cond;
  	void *data;
  	int remove;
};
static int connection_tags = 0;

int ext_yahoo_add_handler (int id, int fd, yahoo_input_condition cond, void *data)
{
	struct _conn *c = y_new0 (struct _conn, 1);
  	c->tag = ++connection_tags;
  	c->id = id;
  	c->fd = fd;
  	c->cond = cond;
  	c->data = data;
  	connections = y_list_prepend (connections, c);
  	return c->tag;
}

void ext_yahoo_remove_handler (int id, int tag)
{
	YList *l;
  	for (l = connections; l; l = y_list_next (l))
    	{
      		struct _conn *c = l->data;
      		if (c->tag == tag)
		{
	  		/* don't actually remove it, just mark it for removal */
	  		/* we'll remove when we start the next poll cycle */
	  		c->remove = 1;
	  		return;
		}
    	}
}

struct connect_callback_data
{
	yahoo_connect_callback callback;
  	void *callback_data;
  	int id;
  	int tag;
};

void connect_complete (void *data, int source, yahoo_input_condition condition)
{
	struct connect_callback_data *ccd = data;
  	int error, err_size = sizeof (error);
  	ext_yahoo_remove_handler (0, ccd->tag);
  	getsockopt (source, SOL_SOCKET, SO_ERROR, &error, (socklen_t *) & err_size);
  	if (error)
    	{
      		close (source);
      		source = -1;
    	}
  	ccd->callback (source, error, ccd->callback_data);
  	FREE (ccd);
}

void yahoo_callback (struct _conn *c, yahoo_input_condition cond)
{
	int ret = 1;
  	char buff[MSG_BUF] = { 0 };

  	if (c->id < 0)
    	{
      		connect_complete (c->data, c->fd, cond);
    	}
  	else
    	{
      		if (cond & YAHOO_INPUT_READ)
			ret = yahoo_read_ready (c->id, c->fd, c->data);
      		if (ret > 0 && cond & YAHOO_INPUT_WRITE)
			ret = yahoo_write_ready (c->id, c->fd, c->data);

      		if (ret == -1)
			snprintf (buff, sizeof (buff),"Yahoo read error (%d): %s", errno, strerror (errno));
		else if (ret == 0)
			snprintf (buff, sizeof (buff), "Yahoo read error: Server closed socket");

    	}
}

int ext_yahoo_connect_async (int id, const char *host, int port, yahoo_connect_callback callback, void *data)
{
	struct sockaddr_in serv_addr;
  	static struct hostent *server;
  	int servfd;
  	struct connect_callback_data *ccd;
  	int error;

  	if (!(server = gethostbyname (host)))
    	{
      		errno = h_errno;
      		return -1;
    	}

  	if ((servfd = socket (AF_INET, SOCK_STREAM, 0)) < 0)
      		return -1;

 	memset (&serv_addr, 0, sizeof (serv_addr));
  	serv_addr.sin_family = AF_INET;
  	memcpy (&serv_addr.sin_addr.s_addr, *server->h_addr_list, server->h_length);
  	serv_addr.sin_port = htons (port);

  	error = connect (servfd, (struct sockaddr *) &serv_addr, sizeof (serv_addr));
  	if (!error)
    	{
      		callback (servfd, 0, data);
      		return 0;
    	}
  	else if (error == -1 && errno == EINPROGRESS)
    	{
      		ccd = calloc (1, sizeof (struct connect_callback_data));
      		ccd->callback = callback;
      		ccd->callback_data = data;
      		ccd->id = id;
      		ccd->tag = ext_yahoo_add_handler (-1, servfd, YAHOO_INPUT_WRITE, ccd);
      		return ccd->tag;
    	}
	else
    	{		
      		close (servfd);
      		return -1;
    	}
}

/*
 * Callback handling code ends here
 ***********************************/

void process_commands (char *line)
{
	char *cmd = strdup(line), *dup = cmd, *tok;

 	if(cfg.message_mode)
	{
		if(line[0] == '!')
		{
			cmd = cmd + 1;
			goto thisiscmd;
		}
		if(!strncasecmp(line, "qq", strlen("qq")))
		{
			cfg.message_mode = 0;
			goto return_back;
		}
		send_msg(line);
		goto return_back;
	}
 thisiscmd:
 	ymshell_addcmdtohist(cmd);	/* add this command to history */
	tok = strtok(cmd, " \t");
	if(tok == NULL)	goto return_back;
	if(!strcasecmp(tok, "secure"))
	{
	#ifdef SECURE_CHAT
		cfg.secure = 1 - cfg.secure;
		if(cfg.secure)
			printw("Secure chat enabled\n");
		else
			printw("Secure chat disabled\n");
	#else
		printw("Secure chat feature disabled. Please re-compile\n");
	#endif	
		goto return_back;	
	}	
	if(!strcasecmp(tok, "reap"))
	{
		ymshell_remove_context(&cmd[5]);
		goto return_back;
	}	
	if(!strcasecmp(tok, "bell"))
	{
		cfg.notice = 1 - cfg.notice;
		if(cfg.notice)
			printw("Notification enabled\n");
		else	
			printw("Notification disabled\n");
		goto return_back;
	}
	if(!strcasecmp(tok, "talk"))
	{
		beep();
		tok = strtok(0, " \t");
		if(tok && !strcmp(tok, "-i")) {
			tok = strtok(0, " \t");
			tok?ymshell_message_index(tok):({goto talk_help;});
		}
		else {
			tok?ymshell_message_id(tok):({goto talk_help;});
		}
		goto return_back;
	talk_help:
		wattron(stdscr, COLOR_PAIR(5) | A_REVERSE);
		printw("Usage : talk <yahoo_id> | -i <index>\n");
		wattroff(stdscr, COLOR_PAIR(5) | A_REVERSE);
		goto return_back;
	}
	if(!strcasecmp(tok, "info"))
	{
		ymshell_display_help();
		goto return_back;
	}
	else
	if(!strcasecmp(tok, "exit"))
	{
		yahoo_logout();
		goto return_back;
	}
	else
	if(!strcasecmp(tok, "avail"))
	{	
		ymshell_change_state(1);
		goto return_back;
  	}
	if(!strcasecmp(tok, "invis"))
	{
		ymshell_change_state(0);
		goto return_back;
	}	
	else
	if(!strcasecmp(tok, "state"))
	{
		ymshell_show_state();
		goto return_back;
	}
	else
	if(!strcasecmp(tok, "bash"))
	{
		ctrl_z_handler();
		goto return_back;
	}	
	if(!strcasecmp(tok, "sh"))
	{
		ctrl_z_handler();
		goto return_back;
	}	
	if(!strcasecmp(tok, "ps"))
	{
		ymshell_show_context();
		goto return_back;
	}
	else
	if(!strcasecmp(tok, "status"))
	{
		ymshell_set_statemsg(&cmd[7]);
		goto return_back;
	}	
	if(!strcasecmp(tok, "auto"))
	{
		ymshell_set_autoreply(&cmd[5]);
		goto return_back;
	}
  	else
  	if(!strcasecmp(tok, "cls"))
  	{
	  	ymshell_clear();
		goto return_back;
  	}
  	else
	if(!strcasecmp(tok, "ls"))
  	{
		tok = strtok(0, " \t");
		if(tok && !strcmp(tok, "-l")) {
			tok = strtok(0, " \t");
			ymshell_listall_buddies(tok);
		} 
		else if(tok && !strcmp(tok, "-g")) {
			tok = strtok(0, " \t");
			tok?({goto ls_help;}):ymshell_list_groups();
		}
		else {
			ymshell_listonline_buddies(tok);
		}
		goto return_back;
	ls_help:
		wattron(stdscr, COLOR_PAIR(5) | A_REVERSE);
		printw("Usage : ls [([-l] [<group_index] | -g)]\n");
		wattroff(stdscr, COLOR_PAIR(5) | A_REVERSE);
		goto return_back;
  	}
	else
	if(!strcasecmp(tok, "wall"))
	{	
		int i;
		tok = strtok(0, " \t");
		if(tok)
			i = atoi(tok);
		else
			goto wall_help;
		tok = strtok(0, " \t");
		if(tok && !strcmp(tok, "-s")) {
			tok = tok + strlen(tok) + 1;
			tok?ymshell_message_group(i, tok, SPAM):({goto wall_help;});
		} 
		else {
			tok[strlen(tok)] = ' ';
			tok?ymshell_message_group(i, tok, NOSPAM):({goto wall_help;});
		}
		goto return_back;
	wall_help:
		wattron(stdscr, COLOR_PAIR(5) | A_REVERSE);
		printw("Usage : wall <group index> [-s] message\n");
		wattroff(stdscr, COLOR_PAIR(5) | A_REVERSE);
		goto return_back;
  	}
	else
	if(!strcasecmp(tok, "touch"))
	{
		tok = strtok(0, " \t");
		tok?ymshell_add_buddy(tok):({goto touch_help;});
		goto return_back;
	touch_help:
		wattron(stdscr, COLOR_PAIR(5) | A_REVERSE);
		printw("Usage : touch <yahoo_id>\n");
		wattroff(stdscr, COLOR_PAIR(5) | A_REVERSE);
		goto return_back;
	}	
	else
	if(!strcasecmp(tok, "rm"))
	{
		tok = strtok(0, " \t");
		tok?ymshell_remove_buddy(tok):({goto rm_help;});
		goto return_back;
	rm_help:
		wattron(stdscr, COLOR_PAIR(5) | A_REVERSE);
		printw("Usage : rm <yahoo_id>\n");
		wattroff(stdscr, COLOR_PAIR(5) | A_REVERSE);
		goto return_back;
	}	
	else
	{
		ymshell_nosuch_command(cmd);
		goto return_back;
	}	

return_back:
  	FREE (dup);
}


void bkspace_handler()
{
	int x, y;
	getyx(stdscr, y, x);
	wmove(stdscr, y, x-1);
	delch();
	touchwin(stdscr);
	wrefresh(stdscr);
}	

void try_auto_complete()		// only for MY command
{	
	int i, x = 0, hit, first_hit = -1, first_len;
	char tmp[MAX_BUF], hold[MAX_BUF];
	char *ptr = &mbuf[0];
	YGroupList *itr = buddygroups;
	yahoo_account_list *ya_list;
	yahoo_account *ya;

	if(ix < 5)	return;
	if(mbuf[0] == '!')	ptr = &mbuf[1];

	if((strncasecmp(ptr, "talk", 4)) == 0)
	{
		i = 5; 
		if(mbuf[0] == '!')	++i;

		while(i != ix)
			tmp[x++] = mbuf[i++];
		tmp[x] = '\0';
	next_itr:
		hit=0;
		do {
			ya_list = itr->ya_list;
			do {
				ya = ya_list->ya;
				if((strncasecmp(tmp, ya->yahoo_id, strlen(tmp))) == 0)
				{
					++hit;
					strcpy(hold, ya->yahoo_id);
				}
				ya_list=ya_list->prev;
			} while(ya_list!=itr->ya_list);
			itr=itr->prev;
		} while(itr!=buddygroups);
		if(first_hit == -1) {
			first_len = strlen(tmp);
			if(hit ==1)
				goto fill_up_hit;
			first_hit = hit;
		}
		if(first_hit == hit) {
			if(hold[strlen(tmp)+1] != '\0') {
				tmp[strlen(tmp)+1]='\0';
				tmp[strlen(tmp)] = hold[strlen(tmp)];
				itr=buddygroups;
				goto next_itr; 
			}
		}
		hold[strlen(tmp)-1]='\0';
		goto fill_up_hit;
	}
	else
		return;
	
fill_up_hit:
	for(x = first_len; x < strlen(hold); x++)
	{
		printw("%c", hold[x]);
		mbuf[ix++] = hold[x];
	}	
	return;	
}

void tab_handler()
{
	if(cfg.message_mode && mbuf[0] != '!')
	{
		mbuf[ix] = ' '; ++ix;
		mbuf[ix] = ' '; ++ix;
		mbuf[ix] = ' '; ++ix;
		mbuf[ix] = ' '; ++ix;
		printw("    ");
	}
	else if(ix > 0)
		try_auto_complete();
}		

void handle_alt_combo(char ch)
{
	if(tolower(ch) == sc.send_cycle)
	{
		ymcycle_context(send_cl);
		return;
	}	
	else if(tolower(ch) == sc.recv_cycle)
	{
		ymcycle_context(recv_cl);
		return;
	}	

	switch(ch) {
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
		ymswitch_context(ch-1);
		break;
	case '0':
		ymswitch_context(ch+9);
		break;
	}
}

#define COMMAND_END(c, inx, buf)	\
					\
	buf[inx] = '\0';		\
	inx = 0;			\
	goto end;

void wait_for_input()
{
	int ch = 0;
	do
	{
		ch=getch();
		if(ch == ERR)
			return;

		switch(ch)
		{
		case 27:
				ch=getch();
				handle_alt_combo(ch);
				break;

		case KEY_UP:
				if(cfg.message_mode == 0)
					ymshell_gethist(0);
				break;
		case KEY_DOWN:	
			//	ymshell_gethist(1);
				break;
		case CTRL_MAP('q'):
				if(cfg.message_mode)
				{
					cfg.message_mode = 0;
					printw("\n");
					goto exit_out;
				}
				return;
		case CTRL_MAP('c'):	
				yahoo_logout();
				break;
		case CTRL_MAP('z'):	//Ctrl - Z, drops to shell
				ctrl_z_handler();
				break;
		case CTRL_MAP('l'):
				ymshell_clear();
				goto exit_out;
				break;
		case '\t':
				tab_handler();
				break;
		case 127:				
		case CTRL_MAP('h'):		
		case CTRL_MAP('?'):
		case KEY_DC:
		case KEY_BACKSPACE:
				if(ix > 0)
				{
					bkspace_handler();
					ix = ix - 1;
				}	
				break;
		case '\n':
				RESET_ITER;
				printw("%c", ch);
				if(ix == 0)
					goto exit_out;
				COMMAND_END(ch, ix, mbuf);	
		default:
				printw("%c", ch);
				mbuf[ix] = ch;
				++ix;
				break;
		}
		fflush(stdout);
	} while(ch != '\n');
end:
	fflush(stdout);
	process_commands(mbuf);
	drop_prompt();
	return;

exit_out:
	drop_prompt();
	fflush(stdout);
	return;
}

static void register_callbacks ()
{
#ifdef USE_STRUCT_CALLBACKS
  static struct yahoo_callbacks yc;

  yc.ext_yahoo_login_response = ext_yahoo_login_response;
  yc.ext_yahoo_got_buddies = ext_yahoo_got_buddies;
  yc.ext_yahoo_got_ignore = ext_yahoo_got_ignore;
  yc.ext_yahoo_got_identities = ext_yahoo_got_identities;
  yc.ext_yahoo_got_cookies = ext_yahoo_got_cookies;
  yc.ext_yahoo_status_changed = ext_yahoo_status_changed;
  yc.ext_yahoo_got_im = ext_yahoo_got_im;
  yc.ext_yahoo_got_conf_invite = ext_yahoo_got_conf_invite;
  yc.ext_yahoo_conf_userdecline = ext_yahoo_conf_userdecline;
  yc.ext_yahoo_conf_userjoin = ext_yahoo_conf_userjoin;
  yc.ext_yahoo_conf_userleave = ext_yahoo_conf_userleave;
  yc.ext_yahoo_conf_message = ext_yahoo_conf_message;
  yc.ext_yahoo_chat_cat_xml = ext_yahoo_chat_cat_xml;
  yc.ext_yahoo_chat_join = ext_yahoo_chat_join;
  yc.ext_yahoo_chat_userjoin = ext_yahoo_chat_userjoin;
  yc.ext_yahoo_chat_userleave = ext_yahoo_chat_userleave;
  yc.ext_yahoo_chat_message = ext_yahoo_chat_message;
  yc.ext_yahoo_chat_yahoologout = ext_yahoo_chat_yahoologout;
  yc.ext_yahoo_chat_yahooerror = ext_yahoo_chat_yahooerror;
  yc.ext_yahoo_got_webcam_image = ext_yahoo_got_webcam_image;
  yc.ext_yahoo_webcam_invite = ext_yahoo_webcam_invite;
  yc.ext_yahoo_webcam_invite_reply = ext_yahoo_webcam_invite_reply;
  yc.ext_yahoo_webcam_closed = ext_yahoo_webcam_closed;
  yc.ext_yahoo_webcam_viewer = ext_yahoo_webcam_viewer;
  yc.ext_yahoo_webcam_data_request = ext_yahoo_webcam_data_request;
  yc.ext_yahoo_got_file = ext_yahoo_got_file;
  yc.ext_yahoo_contact_added = ext_yahoo_contact_added;
  yc.ext_yahoo_rejected = ext_yahoo_rejected;
  yc.ext_yahoo_typing_notify = ext_yahoo_typing_notify;
  yc.ext_yahoo_game_notify = ext_yahoo_game_notify;
  yc.ext_yahoo_mail_notify = ext_yahoo_mail_notify;
  yc.ext_yahoo_got_search_result = ext_yahoo_got_search_result;
  yc.ext_yahoo_system_message = ext_yahoo_system_message;
  yc.ext_yahoo_error = ext_yahoo_error;
  yc.ext_yahoo_log = ext_yahoo_log;
  yc.ext_yahoo_add_handler = ext_yahoo_add_handler;
  yc.ext_yahoo_remove_handler = ext_yahoo_remove_handler;
  yc.ext_yahoo_connect = ext_yahoo_connect;
  yc.ext_yahoo_connect_async = ext_yahoo_connect_async;

  yahoo_register_callbacks (&yc);

#endif
}

/* initialization routines */

void init_yahoo_connection(yahoo_local_account *ylad)
{
	char *ptr;

	cfg.port_to_use = 5050;
	ylad->yahoo_id[0] = '\0';
	ylad->password[0] = '\0';
	read_yahooconf ();
	printf ("Yahoo Id: ");
	if(ylad->yahoo_id[0] == '\0')
		scanf ("%s", ylad->yahoo_id);
	else
		printf("%s\n", ylad->yahoo_id);
	printf ("Password: ");
	if(ylad->password[0] == '\0')
	{
		ptr = getpass ("");
		strcpy (ylad->password, ptr);
	}
	else
	{
		printf("*******\n");
	}	
	register_callbacks ();
	ext_yahoo_login (ylad);
}	
void dump_exit_stack()
{
	printf("*************************************************\n");
	printf("                 YMSHELL exited		 \n");
	printf("               Functions on stack                \n");
	printf("%x\n", (unsigned int)__builtin_return_address(0));
	printf("%x\n", (unsigned int)__builtin_return_address(1));
	printf("%x\n", (unsigned int)__builtin_return_address(2));
	printf("%x\n", (unsigned int)__builtin_return_address(3));
	printf("*************************************************\n");
	return;
}	
void dump_stack_trace(int sig)
{
	endwin();
	printf("*************************************************\n");
	printf("                 YMSHELL Crashed		 \n");
	printf("               Functions on stack                \n");
	printf("%x\n", (unsigned int)__builtin_return_address(0));
	printf("%x\n", (unsigned int)__builtin_return_address(1));
	printf("%x\n", (unsigned int)__builtin_return_address(2));
	printf("%x\n", (unsigned int)__builtin_return_address(3));
	printf("%x\n", (unsigned int)__builtin_return_address(4));
	printf("%x\n", (unsigned int)__builtin_return_address(5));
	printf("%x\n", (unsigned int)__builtin_return_address(6));
	printf("%x\n", (unsigned int)__builtin_return_address(7));
	printf("              Lets hope it dumps core		 \n");
	printf("*************************************************\n");
	exit(-1);
}	
void init_others()
{
	int i;

	signal(SIGSEGV, dump_stack_trace);
	send_cl=(char **)malloc(sizeof(char *)*num_context);
	recv_cl=(char **)malloc(sizeof(char *)*num_context);
	context=(char **)malloc(sizeof(char *)*num_context);
	for(i=0; i<num_context; ++i) {
		send_cl[i]=(char *)malloc(sizeof(char)*255);
		recv_cl[i]=(char *)malloc(sizeof(char)*255);
		context[i]=(char *)malloc(sizeof(char)*255);
		context[i][0] = recv_cl[i][0] = send_cl[i][0] = '\0';
	}
	logfd = fopen ("/tmp/ymshell.log", "w");	/* logging file */
	cfg.notice = 1;
}	

void init_and_check_kde()
{
	int res;

	/* we need this check, as we may not have KDE or we may
	   run ymshell thru telnet, in both cases we cant do notification */
	res = system("kdialog --passivepopup \"logging in\" 1 > /dev/null 2> /dev/null");
	if(res == 0)
		cfg.kde_pres = 1;
}

void init_term()
{
	initscr();
	cbreak();
	raw();
	noecho();
	scrollok(stdscr,TRUE);
	idlok(stdscr, TRUE);
	nodelay(stdscr, TRUE);
	flushinp();
	keypad(stdscr, TRUE);
	start_color();
	init_pair(1, COLOR_GREEN, COLOR_BLACK);
	init_pair(2, COLOR_MAGENTA, COLOR_BLACK);
	init_pair(3, COLOR_YELLOW, COLOR_BLACK);
	init_pair(4, COLOR_BLACK, COLOR_YELLOW);
	init_pair(5, COLOR_BLACK, COLOR_RED);
	init_pair(6, COLOR_BLACK, COLOR_YELLOW);
	signal(SIGWINCH, SIG_IGN);
}

void init_rest()
{
	init_others();
	init_and_check_kde();
	init_term();
}

void is_first_run()
{
	char *hm = getenv("HOME");
	FILE *fp;

	strcpy(mbuf, hm);
	strcat(mbuf, "/.ymshell.conf");
	fp = fopen(mbuf, "r");
	if(fp != NULL)	
	{
		printf("Using %s as conf file\n", mbuf);
		printf("****************************************\n");
		fflush(fp);
		fclose(fp);
	}
	else
	{
		run_config_wizard(getenv("HOME"));
	}	
	return;
}	

/* main loop/program */
int main (int argc, char *argv[])
{
	int lfd, fd_stdin;
	struct timeval tv;
	fd_set inp, outp;
	YList *l;

	fd_stdin = fileno (stdin);
	if(argc > 1 && (strcasecmp(argv[1], "--help") == 0))
	{
		ymshell_display_help();
		printf("\n\nRun ymshell to login now.\n");
		exit(1);
	}	
	is_first_run();

	cfg.poll_loop = 1;
	ylad = y_new0 (yahoo_local_account, 1);
	init_yahoo_connection(ylad);
	while (cfg.poll_loop)
    	{	
		FD_ZERO (&inp); FD_ZERO (&outp); FD_SET (fd_stdin, &inp);
                tv.tv_sec = 0, tv.tv_usec = 100;
                lfd = 0;

		for (l = connections; l;)
		{
	  		struct _conn *c = l->data;
	  		if (c->remove)
			{
	      			YList *n = y_list_next (l);
	      			connections = y_list_remove_link (connections, l);
	      			y_list_free_1 (l);
	      			free (c);
	      			l = n;
	    		}
	  		else
	    		{
	      			if (c->cond & YAHOO_INPUT_READ)
					FD_SET (c->fd, &inp);
	      			if (c->cond & YAHOO_INPUT_WRITE)
					FD_SET (c->fd, &outp);
	      			if (lfd < c->fd)
					lfd = c->fd;
	      			l = y_list_next (l);
	    		}
		}

      		select (lfd + 1, &inp, &outp, NULL, &tv);
		if(cfg.connected)	 wait_for_input();	
		else		 printf("Connecting . . .\r");

		for (l = connections; l; l = y_list_next (l))
		{
	  		struct _conn *c = l->data;
	  		if (c->remove)
	    			continue;
	  		if (FD_ISSET (c->fd, &inp))
	    			yahoo_callback (c, YAHOO_INPUT_READ);
	  		if (FD_ISSET (c->fd, &outp))
	    			yahoo_callback (c, YAHOO_INPUT_WRITE);
		}
    	}


	if(logfd)	fclose(logfd);
  	while (connections)
    	{
      		YList *tmp = connections;
      		struct _conn *c = connections->data;
      		close (c->fd);
      		FREE (c);
      		connections = y_list_remove_link (connections, connections);
      		y_list_free_1 (tmp);
    	}
	yahoo_logout ();
	FREE (ylad);
	return 0;
}
